# -*- coding: utf-8 -*-
# Copyright © 2007 Lateef Alabi-Oki
#
# This file is part of striim.
#
# striim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# striim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with striim; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

"""
This module documents a class that streams music data from internet
radio stations.

@author: Lateef Alabi-Oki
@organization: The striim Project
@copyright: Copyright © 2007 Lateef Alabi-Oki
@license: GNU GPLv2 or Later
@contact: mystilleef@gmail.com
"""
streamer_dbus_service = "org.sourceforge.striimStreamer"
streamer_dbus_path = "/org/sourceforge/striimStreamer"

class StreamController(object):
	"""
	This class creates an object that communicates with the stream
	processor via D-Bus.
	"""

	def __init__(self, player):
		"""
		Initialize object.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.

		@param player: Reference to the Player.
		@type player: A Player object.
		"""
		self.__init_attributes(player)
		from gobject import idle_add
		idle_add(self.__start_streamer)
		self.__sig_id_1 = player.connect("toggle-play", self.__toggle_play_cb)
		from gnomevfs import monitor_add, MONITOR_FILE
		self.__monitor_id = monitor_add(self.__database_uri, MONITOR_FILE,
										self.__database_changed_cb)
		from Striim.Global import session_bus
		session_bus.add_signal_receiver(self.__name_change_cb,
						'NameOwnerChanged',
						'org.freedesktop.DBus',
						'org.freedesktop.DBus',
						'/org/freedesktop/DBus',
						arg0=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__stream_update_cb,
						signal_name="stream_update",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__stream_error_cb,
						signal_name="stream_error",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__playlist_error_cb,
						signal_name="playlist_error",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__stream_message_cb,
						signal_name="stream_message",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__stream_error_cb,
						signal_name="stream_error",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__ready_cb,
						signal_name="ready",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__stopped_cb,
						signal_name="stopped",
						dbus_interface=streamer_dbus_service)
		session_bus.add_signal_receiver(self.__busy_cb,
						signal_name="busy",
						dbus_interface=streamer_dbus_service)

	def __init_attributes(self, player):
		"""
		Initialize data attributes.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.

		@param player: Reference to the Player.
		@type player: A Player object.
		"""
		self.__player = player
		self.__streamer = None
		from os.path import join, split
		self.__streamer_cwd = split(globals()["__file__"])[0]
		self.__streamer_executable = join(self.__streamer_cwd, "striimstreamer.py")
		from Striim.Global import metadata_folder
		self.__database_path = metadata_folder + "/playlist.gdb"
		from gnomevfs import get_uri_from_local_path
		self.__database_uri = get_uri_from_local_path(self.__database_path)
		from sys import prefix
		self.__python_executable = prefix + "/bin" + "/python"
		self.__status_id = None
		self.__is_toggling_play = False
		self.__is_changing_volume = False
		self.__feedback = player.feedback
		self.__striim_server_process_id = None
		self.__quiting = False
		self.__sig_id_1 = self.__sig_id_2 = None
		self.__monitor_id = None
		return

	def __start_streamer(self):
		"""
		Start the stream processor.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		try:
			from dbus import DBusException
			from Striim.Global import dbus_iface, session_bus, python_path
			services = dbus_iface.ListNames()
			from operator import contains
			self.__streamer = None
			if contains(services, streamer_dbus_service):
				self.__streamer = session_bus.get_object(streamer_dbus_service, streamer_dbus_path)
			else:
				from gobject import spawn_async
				data = spawn_async([self.__python_executable, self.__streamer_executable, python_path], working_directory=self.__streamer_cwd)
				self.__striim_server_process_id = data[0]
		except DBusException:
			pass
		return False

	def __kill_striim_server(self):
		try:
			from dbus import DBusException
			from Striim.Global import dbus_iface, session_bus, python_path
			services = dbus_iface.ListNames()
			from operator import contains, is_
			if contains(services, streamer_dbus_service):
				from os import kill
				if is_(self.__striim_server_process_id, None): return
				from signal import SIGHUP
				kill(self.__striim_server_process_id, SIGHUP)
		except DBusException:
			pass
		return

########################################################################

	def __name_change_cb(self, *args):
		"""
		Handles callback when the stream processor quits or starts.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		if self.__quiting: return
		from Striim.Global import session_bus
		try:
			self.__streamer = session_bus.get_object(streamer_dbus_service, streamer_dbus_path)
		except:
			self.__streamer = None
		if self.__streamer: return
		from gobject import idle_add
		idle_add(self.__start_streamer)
		return

	def __ready_cb(self, *args):
		"""
		Handles callback when the stream processor is ready.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		self.__player.emit("streamer-is-ready")
		return

	def __stopped_cb(self, *args):
		"""
		Handles callback when the stream processor is stopped.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		self.__player.emit("streamer-is-stopped")
		if self.__status_id: self.__feedback.unset_modal_message(self.__status_id)
		self.__status_id = None
		return

	def __stream_update_cb(self, metadata_dictionary):
		"""
		Handles callback when metadata information is updated.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.

		@param metadata_dictionary: Information about the track playing.
		@type metadata_dictionary: A Dictionary object.
		"""
		self.__player.response()
		self.__player.emit("stream-update", metadata_dictionary)
		message = "<i>%s</i>" % metadata_dictionary["title"]
		if self.__status_id: self.__feedback.unset_modal_message(self.__status_id)
		self.__status_id = self.__feedback.set_modal_message(message, "info")
		self.__player.response()
		return

	def __stream_message_cb(self, message):
		"""
		Handles callback when stream is buffering.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.

		@param message: Buffering status.
		@type message: A String object.
		"""
		self.__player.response()
		self.__player.emit("stream-message", message)
		from operator import eq
		if eq(message, "done"):
			self.__feedback.reset()
		else:
			message = "<span color='brown'>%s</span>" % message
			self.__feedback.update_status_message(message, "run", 600)
			self.__player.response()
		return

	def __stream_error_cb(self, error):
		"""
		Handles callback when a stream error occurs.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.

		@param error: An error message.
		@type error: A String object.
		"""
		self.__player.response()
		self.__player.emit("stream-error", error)
		data = "Error! " + error.strip(".")
		message = "<b><span color='red'>%s</span></b>" % data
		if self.__status_id: self.__feedback.unset_modal_message(self.__status_id)
		self.__status_id = self.__feedback.set_modal_message(message, "error")
		self.__player.response()
		return

	def __playlist_error_cb(self, *args):
		"""
		Handles callback when no radio station is found.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		self.__player.response()
		self.__player.emit("playlist-error")
		if self.__status_id: self.__feedback.unset_modal_message(self.__status_id)
		from i18n import msg0001
		self.__status_id = self.__feedback.set_modal_message(msg0001, "warning")
		self.__player.response()
		return

	def __busy_cb(self, *args):
		"""
		Handles callback when stream processor is getting metadata.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		self.__player.response()
		self.__player.emit("streamer-is-busy")
		self.__feedback.unset_modal_message(self.__status_id)
		from Striim.i18n import msg0001
		if self.__status_id: self.__feedback.unset_modal_message(self.__status_id)
		self.__status_id = self.__feedback.set_modal_message(msg0001, "connect")
		self.__player.response()
		return

	def __toggle_play_cb(self, *args):
		"""
		Handles callback to play/pause striim.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		from operator import not_
		if not_(self.__streamer): return self.__start_streamer()
		self.__streamer.toggle_play(dbus_interface=streamer_dbus_service,
					reply_handler=self.__reply_handler_cb,
					error_handler=self.__error_handler_cb)
		return

	def __play_new_station(self):
		from operator import not_
		if not_(self.__streamer): return self.__start_streamer()
		self.__streamer.play_new_station(dbus_interface=streamer_dbus_service,
					reply_handler=self.__reply_handler_cb,
					error_handler=self.__error_handler_cb)
		return

	def __reply_handler_cb(self):
		return

	def __error_handler_cb(self, error):
		self.__kill_striim_server()
		print "========================================================"
		print error
		print "========================================================"
		print "RESTARTING STRIIM SERVER"
		return

	def __database_changed_cb(self, monitor_uri, info_uri, event_type):
		"""
		Handles callback when the playlist database is modified.

		@param self: Reference to the Manager instance.
		@type self: An Manager object.

		@param monitor_uri: The uri that is monitored.
		@type monitor_uri: A String object.

		@param info_uri: The uri that is monitored.
		@type info_uri: A String object.

		@param event_type: The type of modification that occured.
		@type event_type: A gnomevfs.MONITOR_EVENT* object.
		"""
		from gnomevfs import MONITOR_EVENT_DELETED
		from gnomevfs import MONITOR_EVENT_CREATED
		from gnomevfs import MONITOR_EVENT_CHANGED
		if event_type in [MONITOR_EVENT_DELETED, MONITOR_EVENT_CREATED, MONITOR_EVENT_CHANGED]:
			self.__play_new_station()
		return

	def destroy(self):
		"""
		Destroy object.

		@param self: Reference to the StreamController instance.
		@type self: A StreamController object.
		"""
		self.__quiting = True
		self.__kill_striim_server()
		from Striim.utils import delete_attributes, disconnect_signal
		disconnect_signal(self.__sig_id_1, self.__player)
		disconnect_signal(self.__sig_id_2, self.__player)
		delete_attributes(self)
		del self
		self = None
		return
